unit TTwitterBot2ClassUnit;

interface

uses
  System.SysUtils, System.Classes, System.DateUtils, System.StrUtils,
  System.Types,
  // Indy
  IdBaseComponent, IdComponent, IdTCPConnection, IdTCPClient, IdHTTP,
  IdSSLOpenSSL, IdHMACSHA1, IdGlobal, IdCoderMIME, IdURI, IdLogFile,
  IdCoder, IdMultipartFormData, IdCompressorZLib,

  System.IOUtils, System.JSON, System.NetEncoding, FMX.Dialogs,
  System.Hash, System.Net.URLClient, System.JSON.Serializers, System.JSON.Converters;

type
  TAuthState = (osNoneAuthorized, osInProcess, osAuthorized);
  THTTPProc = (hpNone, hpAUth_v1, hpAuth_v2);
  THTTPMethod = (hmGet, hmPost);
  TAPIVersion = (av11, av20);

  TBrowserFormShowEvent = procedure (Url: string; Hint: string) of object;
  TUrlChangedEvent = procedure (NewUrl: string; var DoCloseWebForm : boolean) of object;
  TResultEvent = reference to procedure (const AResult: string; const AuthStateChanged: boolean);
  TAuthStateChangedEvent = procedure (AuthState: TAuthState; APIVersion: TAPIVersion) of object;
  TSetJsonEvent = procedure (Json: string) of object;

  TMessageRec = record
    Name: string;
    Text: string;
  end;

  TTwitterBot2 = class
  strict private
    // API v1.1
    v1_ConsumerKey: string;
    v1_ConsumerKeySecret: string;
    v1_OAuthtoken: string;
    v1_OAuthtokenSecret: string;
    v1_RefreshToken: string;
    v1_AccessToken: string;          // obtained by auth
    v1_AccessTokenSecret: string;    // obtained by auth
    v1_Verifier: string;             // PIN
    v1_CallBack: string;
    v1_AuthState: TAuthState;
    v1_Nonce: string;
    v1_TimeStamp: string;
    v1_Status: string;

    // API v2.0
    v2_ClientID: string;
    v2_ClientSecret: string;
    v2_Scopes: string;
    v2_RedirectURL: string;
    v2_ChallengeMethod: string;
    v2_CodeVerifier: string;
    v2_Challenge: string;
    v2_RefreshToken: string;         // obtained by auth
    v2_BearerToken: string;          // obtained by auth
    v2_RefreshedAt: TDateTime;
    v2_AuthState: TAuthState;
    v2_State: string;

    IdHTTP: TIdHTTP;
    SSLIO: TIdSSLIOHandlerSocketOpenSSL;
    Log: TIdLogFile;
    FHttpError: boolean;
    FHttpFirstAccess: boolean;

    FHTTPProc: THTTPProc;
    FAuthStateChanged: TAuthStateChangedEvent;

    FBrowserFormShow: TBrowserFormShowEvent;
    FWebBrowserURL: string;
    FUrlChanged: TUrlChangedEvent;

    FSetJsonEvent_v1: TSetJsonEvent;
    FSetJsonEvent_v2: TSetJsonEvent;

    procedure CreateHTTP;
    procedure ResetHTTP;

    // v1
    procedure Set_v1_Tokens(Json: string);
    function Get_v1_TokensAsJson: string;
    function CreateSignature(Method: THTTPMethod; URL: string): string;
    // upload a image file
    function DoUploadImage_v1(Filename: string): string;

    // v2
    procedure Set_v2_Tokens(Json: string);
    function Get_v2_TokensAsJson: string;
    function CreateAuhorizeUrl: string;
    function ParseTokensFromOrgJson(var Json: string; DateTime: TDateTime): boolean;
    // refresh an expired access token (a access token expires after 7200 second)
    function DoRefreshTokens_v2: string;
    // post text with MediaID of Image
    function DoPostTweet_v2(Text: string; MediaID: string): string;

  public
    constructor Create;
    destructor Destroy; override;

    // events
    procedure BrowserUrlChanged(NewUrl: string; var DoCloseWebForm : boolean);  // by BrowserForm
    property UrlChanged: TUrlChangedEvent write FUrlChanged;                    // to MainForm
    property AuthStateChanged: TAuthStateChangedEvent write FAuthStateChanged;  // to MainForm

    // v1
    property v1_Tokens: string read Get_v1_TokensAsJson write Set_v1_Tokens;
    property SetJsonEvent_v1: TSetJsonEvent write FSetJsonEvent_v1;
    property AuthState_v1: TAuthState read v1_AuthState;
    // v1 authentication starts
    procedure Authorize_v1;                                                     // v1 Auth
    // get an acerss token by PIN
    procedure GetAccessToken_v1(Verifier: string);                              // v1 Auth

    // v2
    property v2_Tokens: string read Get_v2_TokensAsJson write Set_v2_Tokens;
    property SetJsonEvent_v2: TSetJsonEvent write FSetJsonEvent_v2;
    property AuthState_v2: TAuthState read v2_AuthState;
    // v2 authentication starts
    procedure Authorize_v2;                                                     // v2 Auth
    // get an acerss token and a refresh token by the Code including in URL
    function GetAccessToken_v2(Url: string): string;

    // Post a tweet with image
    procedure PostTweet(Text: string; FileName: string; AResultEvent: TResultEvent); // Text: v2 - Image: v1

    // tools
    property BrowserFormShow: TBrowserFormShowEvent write FBrowserFormShow;
    procedure BrowserNavigate(URL: string);

  end;

 var
   TwitterBot2: TTwitterBot2 = nil;

implementation

const
  // v1
  RequestTokenURL = 'https://api.twitter.com/oauth/request_token';
  RequestVerifierURL = 'https://twitter.com/oauth/authorize?oauth_token=';
  RequestAccessTokenURL = 'https://api.twitter.com/oauth/access_token';
  UploadMediaURL = 'https://upload.twitter.com/1.1/media/upload.json';
  PostTweetURL_v1 = 'https://api.twitter.com/1.1/statuses/update.json';
  // v2
  AuthorizeURL = 'https://twitter.com/i/oauth2/authorize';
  GetAccessTokenURL = 'https://api.twitter.com/2/oauth2/token';
  PostTweetUrl_v2 = 'https://api.twitter.com/2/tweets';
  AboutMeUrl = 'https://api.twitter.com/2/users/me';

  AuthStateStr: array[TAuthState] of string = ('Start', 'InProcess', 'Authorized');
  HTTPMethodStr: array[THTTPMethod] of string = ('GET', 'POST');

  BooleanStr: array[boolean] of string = ('false', 'true');

const
  UrlEncodeCharSet = ['A'..'Z', 'a'..'z', '0'..'9', '-', '.', '_', '~'];

// percent encode
function UrlEncode(Text: string): string;
var
  Ch: Char;
begin
  result := '';

  try
    for ch in Text do
      if CharInSet(Ch, UrlEncodeCharSet) then
        result := result + Char(Ch)
      else
        result := result + '%'+ IntToHex(ord(Ch), 2);
  except

  end;
end;

function URLEncodeW(Text: string): string;
var
  S: Utf8String;
  Ch: AnsiChar;
begin
  Result := '';
  S := UTF8Encode(Text);

  try
    for Ch in S do
      if CharInSet(Ch, UrlEncodeCharSet) then
        result := result + WideChar(Ch)
      else
        result := result + '%' + IntToHex(Ord(Ch), 2);
  except

  end;
end;

// add a Key and Value pair to a Parameter string (for header)
procedure AddParam(var Param: string; Key, Value: string);
begin
  if Param = '' then
    Param := Key + '=' + '"' + Value + '"'
  else
    Param := Param + ',' + Key + '=' + '"' + Value + '"';
end;

function GetRandomString(ALength: integer): string;
const
  ACharSequence = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890';
var
  Len: integer;
begin
  Len := Length(ACharSequence);
  SetLength(result, ALength);
  Randomize;
  for var i := Low(result) to High(result) do
    result[i] := ACharSequence.Chars[Random(Len)]
end;

function GetBase64_SHA256(Text: string): string;
var
  Bytes: TBytes;
begin
  Bytes := THashSHA2.GetHashBytes(Text, SHA256);
  result := TNetEncoding.Base64.EncodeBytesToString(Bytes);
  result := result.Split(['='])[0];
  result := result.Replace('+', '-');
  result := result.Replace('/', '_');

end;

function MemoryStreamToString(Stream: TMemoryStream): string;
var
  SList: TStringList;
begin
  result := '';
  if Assigned(Stream) and (Stream.size > 0) then begin
    SList := TStringList.Create;
    try
      try
        Stream.Seek(0, TSeekOrigin.soBeginning);
        SList.LoadFromStream(Stream, TEncoding.UTF8);
        result := SList.Text;
      except

      end;
    finally
      SList.Free;
    end;
  end;
end;

// create signature
function TTwitterBot2.CreateSignature(Method: THTTPMethod; URL: string): string;
  // generate 32 bytes Nonce
  function CreateNonce: string;
  var
    i: integer;
  const
    Str: string = '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ';
  begin
    Randomize;
    for i := 0 to 31 do
      result := result + Str[Random(62) + 1];
  end;
  // calculate the signature by SHA1
  function CalcHMACSha1(SignatureBase, SigningKey: string): string;
  var
    SHA1: TIdHMACSHA1;
  begin
    result := '';
    SHA1 := TIdHMACSHA1.Create;
    try
      SHA1.Key := ToBytes(SigningKey);
      result := TIdEncoderMIME.EncodeBytes(SHA1.HashValue(ToBytes(SignatureBase)));
    finally
      SHA1.Free;
    end;
  end;
  // add a Key and Value pair to a Parameter string
  procedure AddParam(var Param: string; Key, Value: string);
  begin
    if Value <> '' then begin
      if Param = '' then
        Param := Key + '=' + Value
      else
        Param := Param + '&' + Key + '=' + Value;
    end;
  end;
var
  SignatureBase: string;
  SigningKey: string;
  Params: string;
begin
  result := '';

  try
    v1_Nonce := CreateNonce;
    v1_TimeStamp := IntToStr(DateTimeToUnix(Now, true)); //CreateTimeStamp;
    v1_Status := UrlEncodeW(v1_Status);

    Params := '';
    //if FCallBack <> '' then
    AddParam(Params, 'oauth_callback', v1_CallBack);    // include_entities
    AddParam(Params, 'oauth_consumer_key', URLEncode(v1_ConsumerKey));
    AddParam(Params, 'oauth_nonce', v1_Nonce);
    AddParam(Params, 'oauth_signature_method', 'HMAC-SHA1');
    AddParam(Params, 'oauth_timestamp', v1_TimeStamp);
    //if FOAuthToken <> '' then
    AddParam(Params, 'oauth_token', URLEncode(v1_OAuthToken));
    //if FVerifier <> '' then
    AddParam(Params, 'oauth_verifier', v1_Verifier);
    AddParam(Params, 'oauth_version', '1.0');
    //if FStatus <> '' then
    AddParam(Params, 'status', v1_Status);

    SignatureBase := HTTPMethodStr[Method] +
                     '&' + URLEncode(URL) +
                     '&' + URLEncode(Params);

    SigningKey := URLEncode(v1_ConsumerKeySecret) +
            '&' + URLEncode(v1_OAuthTokenSecret);

    result := UrlEncode(CalcHMACSha1(SignatureBase, SigningKey));

  except

  end;
end;



{ TTwitterBot2 }

constructor TTwitterBot2.Create;
begin
  CreateHTTP;
  FBrowserFormShow := nil;
  FHTTPProc := THTTPProc.hpNone;

  FUrlChanged := nil;
  FAuthStateChanged := nil;
  FSetJsonEvent_v1 := nil;
  FSetJsonEvent_v2 := nil;

  FHttpError := false;
  FHttpFirstAccess := true;
  FHTTPProc := THTTPProc.hpNone;
end;

destructor TTwitterBot2.Destroy;
begin
  Log.Close;
  Log.Free;
  SSLIO.Free;
  IdHTTP.Free;

  inherited;
end;

procedure TTwitterBot2.Set_v1_Tokens(Json: string);
var
  JSONObject: TJSONObject;
  JsonValue: TJsonValue;
begin
  JSONObject := nil;
  if not Json.IsEmpty then begin
    JSONObject := TJSONObject.ParseJSONValue(Json) as TJSONObject;
    if JSONObject <> nil then begin
      try
        try
          if JSONObject.TryGetValue('consumer_key', JsonValue) then
            v1_ConsumerKey := JsonValue.Value;
          if JSONObject.TryGetValue('consumer_key_secret', JsonValue) then
            v1_ConsumerKeySecret := JsonValue.Value;

          if JSONObject.TryGetValue('bearer_token', JsonValue) then
            v1_RefreshToken := JsonValue.Value;

          if JSONObject.TryGetValue('access_token', JsonValue) then begin
            v1_AccessToken := JsonValue.Value;
            v1_OAuthToken := v1_AccessToken;
          end;
          if JSONObject.TryGetValue('access_token_secret', JsonValue) then begin
            v1_AccessTokenSecret := JsonValue.Value;
            v1_OAuthTokenSecret := v1_AccessTokenSecret;
          end;

          if not v1_AccessToken.IsEmpty and not v1_AccessTokenSecret.IsEmpty then
            v1_AuthState := TAuthState.osAuthorized
          else
            v1_AuthState := TAuthState.osNoneAuthorized;
          if Assigned(FAuthStateChanged) then
            FAuthStateChanged(v1_AuthState, TAPIVersion.av11);
        except

        end;
      finally
        JSONObject.Free;
      end;
    end;
  end;
end;

function TTwitterBot2.Get_v1_TokensAsJson: string;
var
  JSON: TJSONObject;
begin
  result := '';

  JSON := TJSONObject.Create;
  try
    try
      JSON.AddPair('consumer_key', v1_ConsumerKey);
      JSON.AddPair('consumer_key_secret', v1_ConsumerKeySecret);

      JSON.AddPair('bearer_token', v1_RefreshToken);

      JSON.AddPair('access_token', v1_AccessToken);
      JSON.AddPair('access_token_secret', v1_AccessTokenSecret);
      result := Json.ToString;
    except

    end;
  finally
    JSON.Free;
  end;
end;

procedure TTwitterBot2.Set_v2_Tokens(Json: string);
var
  JSONObject: TJSONObject;
  JsonValue: TJsonValue;
begin
  JSONObject := nil;
  if not Json.IsEmpty then begin
    JSONObject := TJSONObject.ParseJSONValue(Json) as TJSONObject;
    if JSONObject <> nil then begin
      try
        try
          if JSONObject.TryGetValue('client_id', JsonValue) then
            v2_ClientID := JsonValue.Value;
          if JSONObject.TryGetValue('client_id_secret', JsonValue) then
            v2_ClientSecret := JsonValue.Value;

          if JSONObject.TryGetValue('scopes', JsonValue) then
            v2_Scopes := JsonValue.Value;
          if JSONObject.TryGetValue('redirect_url', JsonValue) then
            v2_RedirectURL := JsonValue.Value;
          if JSONObject.TryGetValue('challenge_method', JsonValue) then
            v2_ChallengeMethod := JsonValue.Value;

          if JSONObject.TryGetValue('refresh_token', JsonValue) then
            v2_RefreshToken := JsonValue.Value;
          if JSONObject.TryGetValue('bearer_token', JsonValue) then
            v2_BearerToken := JsonValue.Value;
          v2_RefreshedAt := 0;
          if JSONObject.TryGetValue('bearer_token_obtained_at', JsonValue) then begin
            if not JsonValue.Value.IsEmpty then
              v2_RefreshedAt := ISO8601ToDate(JsonValue.Value, false);
          end;

          // IncSecond(v2_RefreshedAt, 2700) > IncMinute(Now, 1)
          if not v2_RefreshToken.IsEmpty and not v2_BearerToken.IsEmpty then
            v2_AuthState := TAuthState.osAuthorized
          else
            v2_AuthState := TAuthState.osNoneAuthorized;

          if Assigned(FAuthStateChanged) then
            FAuthStateChanged(v2_AuthState, TAPIVersion.av20);
        except

        end;
      finally
        JSONObject.Free;
      end;
    end;
  end;

end;

function TTwitterBot2.Get_v2_TokensAsJson: string;
var
  JSON: TJSONObject;
begin
  result := '';

  JSON := TJSONObject.Create;
  try
    try
      JSON.AddPair('client_id', v2_ClientID);
      JSON.AddPair('client_id_secret', v2_ClientSecret);

      JSON.AddPair('scopes', v2_Scopes);
      JSON.AddPair('redirect_url', v2_RedirectURL);
      JSON.AddPair('challenge_method', v2_ChallengeMethod);

      JSON.AddPair('refresh_token', v2_RefreshToken);
      JSON.AddPair('bearer_token', v2_BearerToken);
      JSON.AddPair('bearer_token_obtained_at', DateToISO8601(v2_RefreshedAt, false));

      result := Json.ToString;
    except

    end;
  finally
    JSON.Free;
  end;

end;

procedure TTwitterBot2.CreateHTTP;
begin
  try
    IdHTTP := TIdHTTP.Create(nil);
    IdHTTP.ReadTimeout := 20000;
    IdHTTP.ConnectTimeout := 20000;
    IdHTTP.HandleRedirects := true;
    IdHTTP.HTTPOptions := IdHTTP.HTTPOptions - [hoForceEncodeParams];
    // IdHTTP.HTTPOptions := IdHTTP.HTTPOptions + [hoNoProtocolErrorException, hoWantProtocolErrorContent];
    IdHTTP.Request.UserAgent := 'TTwitterBot v2.0';
    IdHTTP.Request.Accept := '*/*';
    {$IFDEF MSWINDOWS}
    IdHTTP.Compressor := TIdCompressorZLib.Create(IdHTTP);
    {$ENDIF}
    IdHTTP.Request.AcceptEncoding := 'gzip';

    SSLIO := TIdSSLIOHandlerSocketOpenSSL.Create(nil);
    SSLIO.SSLOptions.SSLVersions := [sslvTLSv1_2];
    IdHTTP.IOHandler := SSLIO;

    if not Assigned(Log) then begin
      Log := TIdLogFile.Create(nil);
      Log.Filename :='HttpLog.txt';
    end;
    IdHTTP.Intercept := Log;
    Log.Active := true;

    IdHttp.AllowCookies := true;
  except

  end;
end;

procedure TTwitterBot2.ResetHTTP;
begin
  SSLIO.Free;
  IdHTTP.Free;
  Log.Active := false;

  CreateHTTP;
end;

// events

procedure TTwitterBot2.BrowserUrlChanged(NewUrl: string; var DoCloseWebForm : boolean);
begin
  DoCloseWebForm := false;

  if FHTTPProc = THTTPProc.hpAuth_v1 then begin   // v1

  end else begin                                  // v2
    if Assigned(FUrlChanged) then
      FUrlChanged(NewUrl, DoCloseWebForm);
    if (v2_AuthState = TAuthState.osInProcess) and NewUrl.Contains('code=') then begin
      FWebBrowserURL := NewUrl;
    end else
      DoCloseWebForm := false;
  end;

end;


// Misc

// public
procedure TTwitterBot2.BrowserNavigate(URL: string);
begin
  if Assigned(FBrowserFormShow) then
    FBrowserFormShow(Url, '');

end;

procedure TTwitterBot2.PostTweet(Text: string; FileName: string; AResultEvent: TResultEvent);
var
  Refreshed: boolean;
begin
  FHTTPProc := THTTPProc.hpAuth_v2;

  if not Text.IsEmpty and Assigned(AResultEvent) then begin
    TThread.CreateAnonymousThread(procedure
    var
      MediaID, Response: string;
    begin
      if IncSecond(v2_RefreshedAt, 7000) < Now then begin // 7200
        DoRefreshTokens_v2;
        Refreshed := true;
      end else
        Refreshed := false;

      if (v1_AuthState = TAuthState.osAuthorized) and not FileName.IsEmpty then
        MediaID := DoUploadImage_v1(FileName)
      else
        MediaID := '';

      if not Text.IsEmpty then
        Response := DoPostTweet_v2(Text, MediaID)
      else
        Response := MediaID;

      TThread.Synchronize(
        TThread.CurrentThread,
        procedure
        begin
          AResultEvent(Response, false);
          if Assigned(FSetJsonEvent_v2) and Refreshed then
            FSetJsonEvent_v2(Get_v2_TokensAsJson);
        end);
    end).Start;
  end;
end;

// ---------------------------------------------
//     v1
// ---------------------------------------------

// obtain a Request Token
procedure TTwitterBot2.Authorize_v1;
var
  Params: string;
  Signature: string;
  Body: TStringList;
  Response: string;
  ResponseList: TStringList;
begin
  FHttpError := false;
  FHTTPProc := THTTPProc.hpAuth_v1;

  v1_OAuthToken := '';
  v1_OAuthTokenSecret := '';
  v1_AccessToken := '';
  v1_AccessTokenSecret := '';
  v1_CallBack :=  'oob';
  v1_Status := '';
  v1_Verifier := '';

  Signature := CreateSignature(hmPost, RequestTokenURL);

  Params := '';
  AddParam(Params, 'oauth_callback', v1_CallBack);
  AddParam(Params, 'oauth_consumer_key', v1_ConsumerKey);
  AddParam(Params, 'oauth_nonce', v1_Nonce);
  AddParam(Params, 'oauth_signature', Signature);
  AddParam(Params, 'oauth_signature_method', 'HMAC-SHA1');
  AddParam(Params, 'oauth_timestamp', v1_TimeStamp);
  AddParam(Params, 'oauth_version', '1.0');

  Params := 'OAuth ' + Params;

  // IdHTTP
  IdHTTP.Request.CustomHeaders.Clear;
  IdHTTP.Request.BasicAuthentication := false;
  // IdHTTP.Request.CustomHeaders.FoldLines := false;
  IdHTTP.Request.CustomHeaders.AddValue('Authorization', Params);

  Body := TStringList.Create;
  try
    try
      Response := IdHTTP.Post(RequestTokenURL, Body);

      // parse Response
      ResponseList := TStringList.Create;
      try
        ResponseList.Text := ReplaceText(Response, '&', sLineBreak);
        v1_OAuthToken := ResponseList.Values['oauth_token'];
        v1_OAuthTokenSecret := ResponseList.Values['oauth_token_secret'];
      finally
        ResponseList.Free;
      end;

      if v1_OAuthToken.IsEmpty or v1_OAuthTokenSecret.IsEmpty then begin
        v1_OAuthToken := '';
        v1_OAuthTokenSecret := '';
      end else begin
        if Assigned(FBrowserFormShow) then  // open Browser
          FBrowserFormShow(RequestVerifierURL + v1_OAuthToken, 'Login and authorize a app');
        v1_AuthState := TAuthState.osInProcess;
        if Assigned(FAuthStateChanged) then
          FAuthStateChanged(v1_AuthState, TAPIVersion.av11);
      end;

    except
      on e: exception do begin
        TFile.AppendAllText('HttpError.log',
          FormatDateTime('yyyy-mm-dd hh:nn:ss - ', Now) +
          'TTwitterBot2.Authorize_v1: ' + e.Message + sLineBreak,
          TEncoding.UTF8);
        FHttpError := true;
      end;
    end;
  finally
    Body.Free;
    if IdHTTP.Connected then
      IdHTTP.Disconnect;
  end;

  if FHttpError then
    ResetHTTP;

end;

procedure TTwitterBot2.GetAccessToken_v1(Verifier: string);
var
  Params: string;
  Signature: string;
  Body: TStringList;
  Response: string;
  ResponseList: TStringList;
begin
  FHttpError := false;

  v1_AccessToken := '';
  v1_AccessTokenSecret := '';
  v1_CallBack := '';
  v1_Status := '';
  v1_Verifier := Verifier;

  Signature := CreateSignature(hmPost, RequestAccessTokenURL);

  Params := '';
  AddParam(Params, 'oauth_consumer_key', v1_ConsumerKey);
  AddParam(Params, 'oauth_nonce', v1_Nonce);
  AddParam(Params, 'oauth_signature', Signature);
  AddParam(Params, 'oauth_signature_method', 'HMAC-SHA1');
  AddParam(Params, 'oauth_timestamp', v1_TimeStamp);
  AddParam(Params, 'oauth_token', URLEncode(v1_OAuthToken));
  AddParam(Params, 'oauth_version', '1.0');

  Params := 'OAuth ' + Params;

  IdHTTP.Request.CustomHeaders.Clear;
  IdHTTP.Request.BasicAuthentication := false;
  // IdHTTP.Request.CustomHeaders.FoldLines := false;
  IdHTTP.Request.CustomHeaders.AddValue('Authorization', Params);

  Body := TStringList.Create;
  try
    try
      Body.Add('oauth_verifier=' + v1_Verifier);

      Response := IdHTTP.Post(RequestAccessTokenURL, Body);

      // parse Response
      ResponseList := TStringList.Create;
      try
        ResponseList.Text := ReplaceText(Response, '&', sLineBreak);
        v1_OAuthToken := ResponseList.Values['oauth_token'];;
        v1_OAuthTokenSecret := ResponseList.Values['oauth_token_secret'];
      finally
        ResponseList.Free;
      end;

     if not v1_OAuthToken.IsEmpty and not v1_OAuthTokenSecret.IsEmpty then begin
      v1_AccessToken := v1_OAuthToken;
      v1_AccessTokenSecret := v1_OAuthTokenSecret;
      if Assigned(FBrowserFormShow) then // close Browser
        FBrowserFormShow('', '');
      v1_AuthState := TAuthState.osAuthorized;
      if Assigned(FAuthStateChanged) then
        FAuthStateChanged(v1_AuthState, TAPIVersion.av11);
      if Assigned(FSetJsonEvent_v1) then
        FSetJsonEvent_v1(Get_v1_TokensAsJson);
      end;

    except
      on e: exception do begin
        TFile.AppendAllText('HttpError.log',
          FormatDateTime('yyyy-mm-dd hh:nn:ss - ', Now) +
          'TTwitterBot2.GetAccessToken_v1: ' + e.Message + sLineBreak,
          TEncoding.UTF8);
        FHttpError := true;
      end;
    end;
  finally
    Body.Free;
    if IdHTTP.Connected then
      IdHTTP.Disconnect;
  end;

  if FHttpError then
    ResetHTTP;

end;

//{
//  "media_id": 710511363345354753,
//  "media_id_string": "710511363345354753",
//  "size": 11065,
//  "expires_after_secs": 86400,
//  "image": {
//    "image_type": "image/jpeg",
//    "w": 800,
//    "h": 320
//  }
//}
function TTwitterBot2.DoUploadImage_v1(Filename: string): string;
  function GetMediaID(Json: string; AName: string): string;
  var
    JSONObject: TJSONObject;
    JsonValue: TJsonValue;
  begin
   result := '';
   if not Json.IsEmpty and not AName.IsEmpty then begin
      JSONObject := TJSONObject.ParseJSONValue(Json) as TJSONObject;
      if JSONObject <> nil then begin
        try
          if JSONObject.TryGetValue(AName, JsonValue) then
              result := JsonValue.Value;
        finally
          JSONObject.Free;
        end;
      end;
    end;
  end;
var
  Params: string;
  Signature: string;
  Body: TIdMultiPartFormDataStream;
  Stream: TMemoryStream;
  Json: string;
  SList: TStringList;
begin
  result := '';

  if not TFile.Exists(Filename) then
    exit;

  FHttpError := false;

  IdHTTP.Request.BasicAuthentication := false;
  IdHTTP.Request.CustomHeaders.Clear;
  IdHTTP.Request.RawHeaders.FoldLines := false;
  IdHTTP.Request.CustomHeaders.FoldLines := false;
  IdHTTP.Request.ContentType := 'multipart/form-data';
  IdHTTP.Request.AcceptEncoding := 'gzip';

  v1_CallBack := '';
  v1_Verifier := '';
  v1_Status := '';

  Signature := CreateSignature(hmPost, UploadMediaURL);

  Params := '';
  AddParam(Params, 'oauth_consumer_key', v1_ConsumerKey);
  AddParam(Params, 'oauth_nonce', v1_Nonce);
  AddParam(Params, 'oauth_signature', Signature);
  AddParam(Params, 'oauth_signature_method', 'HMAC-SHA1');
  AddParam(Params, 'oauth_timestamp', v1_TimeStamp);
  AddParam(Params, 'oauth_token', URLEncode(v1_OAuthToken));
  AddParam(Params, 'oauth_version', '1.0');

  Params := 'OAuth ' + Params;

  IdHTTP.Request.CustomHeaders.Clear;
  IdHTTP.Request.BasicAuthentication := false;
  // IdHTTP.Request.CustomHeaders.FoldLines := false;
  IdHTTP.Request.CustomHeaders.AddValue('Authorization', Params);

  Body := TIdMultiPartFormDataStream.Create;
  Stream := TMemoryStream.Create;

  try
    try
      Body.AddFile('media', Filename, 'application/octet-stream');

      IdHTTP.Post(UploadMediaURL, Body, Stream);

      if Stream.Size > 0 then begin
        Stream.Seek(0, TSeekOrigin.soBeginning);
        SList := TstringList.Create;
        try
          SList.LoadFromStream(Stream, TEncoding.UTF8);
          result := GetMediaID(SList.Text, 'media_id_string');
        finally
          SList.Free;
        end;
      end;

    except
      on e: exception do begin
        TFile.AppendAllText('HttpError.log',
          FormatDateTime('yyyy-mm-dd hh:nn:ss - ', Now) +
          'TTwitterBot2.UploadImage: ' + e.Message + sLineBreak,
          TEncoding.UTF8);
        FHttpError := true;
      end;
    end;
  finally
    Body.Free;
    Stream.Free;
    if IdHTTP.Connected then
      IdHTTP.Disconnect;
  end;

  if FHttpError then
    ResetHTTP;

end;

// ---------------------------------------------
//     v2
// ---------------------------------------------

function TTwitterBot2.CreateAuhorizeUrl: string;
  procedure AddParam(var Param: string; Key, Value: string);
  begin
    if Value <> '' then begin
      if Param = '' then
        Param := Key + '=' + Value
      else
        Param := Param + '&' + Key + '=' + Value;
    end;
  end;
var
  Params: string;
begin
  result := '';

  try
    AddParam(Params, 'response_type', 'code');
    AddParam(Params, 'client_id', v2_ClientId);
    AddParam(Params, 'redirect_uri', v2_RedirectURL);
    AddParam(Params, 'scope', UrlEncode(v2_Scopes));
    AddParam(Params, 'state', v2_State);
    AddParam(Params, 'code_challenge', v2_Challenge);
    AddParam(Params, 'code_challenge_method', v2_ChallengeMethod);

    result := AuthorizeURL + '?' + Params;
  except

  end;
end;

procedure TTwitterBot2.Authorize_v2;
var
  URL: string;
begin
  v2_State := GetRandomString(16);
  v2_CodeVerifier := GetRandomString(64);
  if v2_ChallengeMethod = 'plain' then
    v2_Challenge := v2_CodeVerifier
  else
    v2_Challenge := GetBase64_SHA256(v2_CodeVerifier);

  FHTTPProc := THTTPProc.hpAuth_v2;

  URL := CreateAuhorizeUrl;

  if not URL.IsEmpty then begin
    if Assigned(FBrowserFormShow) then begin // open Browser
      FBrowserFormShow(URL, 'Login and authorize a app');
    end;

    v2_AuthState := TAuthState.osInProcess;
    if Assigned(FAuthStateChanged) then
      FAuthStateChanged(v2_AuthState, TAPIVersion.av20);
  end;
end;

// get Access tokens after your user permitted your app
function TTwitterBot2.GetAccessToken_v2(Url: string): string;
var
  URI: TURI;
  Body: TStringList;
  Stream: TMemoryStream;
begin
  result := '';
  FHttpError := false;

  if Assigned(FBrowserFormShow) then  // close Browser
    FBrowserFormShow('', '');

  URI.Create(URL);

  IdHTTP.Request.CustomHeaders.Clear;

  IdHTTP.Request.BasicAuthentication := true;
  IdHTTP.Request.Username := v2_ClientID;
  IdHTTP.Request.Password := v2_ClientSecret;

  IdHTTP.Request.RawHeaders.FoldLines := false;
  IdHTTP.Request.ContentType := 'application/x-www-form-urlencoded';

  Body := TStringList.Create;
  Stream := TMemoryStream.Create;
  try
    Body.Add('grant_type=authorization_code');
    Body.Add('redirect_uri=' + UrlEncode(v2_RedirectUrl));
    Body.Add('code_verifier=' + v2_CodeVerifier);
    Body.Add('client_id=' + v2_ClientID);
    Body.Add('code=' + Uri.ParameterByName['code']);
    try
      IdHTTP.Post(GetAccessTokenURL, Body, Stream);

      result := MemoryStreamToString(Stream);
      if result.Contains('refresh_token') and result.Contains('access_token') then begin
        if ParseTokensFromOrgJson(result, Now) then begin
          v2_AuthState := TAuthState.osAuthorized;
          if Assigned(FAuthStateChanged) then
            FAuthStateChanged(v2_AuthState, TAPIVersion.av20);
          if Assigned(FSetJsonEvent_v2) then
            FSetJsonEvent_v2(Get_v2_TokensAsJson);
          TFile.WriteAllText('Tokens_v2.txt', result, TEncoding.UTF8);
        end;
      end;
    except
      on e: exception do begin
        TFile.AppendAllText('HttpError.log',
          FormatDateTime('yyyy-mm-dd hh:nn:ss - ', Now) +
          'TTwitterBot2.DoGetAccessToken_v2: ' + e.Message + sLineBreak,
          TEncoding.UTF8);
        FHttpError := true;
      end;
    end;
  finally
    Body.Free;
    Stream.Free;
  end;

  // after a use of a Basic Authentication, TIdHTTP must be reset !
  ResetHTTP;

end;

// {"token_type":"bearer",
// "expires_in":7200,
// "access_token":"VlR...",
// "scope":"tweet.write tweet.read offline.access",
// "refresh_token":"Wm5..."}

function TTwitterBot2.ParseTokensFromOrgJson(var Json: string; DateTime: TDateTime): boolean;
var
  JSONObject: TJSONObject;
  JsonValue: TJsonValue;
begin
  result := false;

  if not Json.Contains('refresh_token') then
    exit;

  JSONObject := nil;
  if not Json.IsEmpty then begin
    JSONObject := TJSONObject.ParseJSONValue(Json) as TJSONObject;
    if JSONObject <> nil then begin
      try
        if JSONObject.TryGetValue('access_token', JsonValue) then
          v2_BearerToken := JsonValue.Value;
        if JSONObject.TryGetValue('refresh_token', JsonValue) then
          v2_RefreshToken := JsonValue.Value;
        if JSONObject.TryGetValue('expires_in', JsonValue) then
          v2_RefreshedAt := now; // 7200
        Json := JSONObject.ToString;
        result := true;
      finally
        JSONObject.Free;
      end;
    end;
  end;

end;

// refresh tokens after a bearer token expired
function TTwitterBot2.DoRefreshTokens_v2: string;
var
  Body: TStringList;
  Stream: TMemoryStream;
begin
  result := '';

  IdHTTP.Request.CustomHeaders.Clear;
  IdHTTP.Request.BasicAuthentication := true;
  IdHTTP.Request.Username := v2_ClientID;
  IdHTTP.Request.Password := v2_ClientSecret;

  IdHTTP.Request.RawHeaders.FoldLines := false;
  IdHTTP.Request.ContentType := 'application/x-www-form-urlencoded';

  Body := TStringList.Create;
  Stream := TMemoryStream.Create;
  try
    Body.Add('refresh_token=' + v2_RefreshToken);
    Body.Add('grant_type=refresh_token');
    Body.Add('client_id=' + v2_ClientID);
    try
      IdHTTP.Post(GetAccessTokenURL, Body, Stream);

      result := MemoryStreamToString(Stream);
      if result.Contains('refresh_token') and result.Contains('access_token') then
        ParseTokensFromOrgJson(result, Now);

    except
      on e: exception do begin
        TFile.AppendAllText('HttpError.log',
          FormatDateTime('yyyy-mm-dd hh:nn:ss - ', Now) +
          'TTwitterBot2.DoRefreshTokens_v2: ' + e.Message + sLineBreak,
          TEncoding.UTF8);
      end;
    end;
  finally
    Body.Free;
    Stream.Free;
  end;

  // after a use of a Basic Authentication, TIdHTTP must be reset !
  ResetHTTP;

end;

function TTwitterBot2.DoPostTweet_v2(Text: string; MediaID: string): string;
type
  TTweetMedia = record
    media_ids: TArray<string>;
  end;
  TTweet = record
    text: string;
  end;
  TTweetwMedia = record
    text: string;
    media: TTweetMedia;
  end;
var
  Body: TStringStream;
  Stream: TMemoryStream;

  Json: string;
  Serial: TJsonSerializer;
  TweetwMedia: TTweetwMedia;
  Tweet: TTweet;
  TweetMedia: TTweetMedia;
begin
  result := '';

  FHttpError := false;

  IdHTTP.Request.AcceptEncoding := '';

  IdHTTP.Request.BasicAuthentication := false;

  IdHTTP.Request.CustomHeaders.Clear;
  IdHTTP.Request.CustomHeaders.FoldLines := false;
  IdHTTP.Request.CustomHeaders.AddValue('Authorization', 'Bearer ' + v2_BearerToken);
  IdHTTP.Request.ContentType := 'application/json';
  IdHTTP.Request.CharSet := 'utf-8';

  // {"text": "Tweeting with media!", "media": {"media_ids": ["1455952740635586573"]}}
  //Json := '"text": "' + Text +'"';
  //if not MediaID.IsEmpty then
  //  Json := Json + ',' + '"media": {"media_ids": ["' + MediaID + '"]}';
  //Json := '{' + Json + '}';

  Json := '';
  Serial := TJsonSerializer.Create;
  try
    try
      if MediaID.IsEmpty then begin     // need ?
        Tweet.text := Text;
        Json := Serial.Serialize<TTweet>(Tweet);
      end else begin
        TweetwMedia.text := Text;
        TweetMedia.media_ids := [MediaID];
        TweetwMedia.media := TweetMedia;
        Json := Serial.Serialize<TTweetwMedia>(TweetwMedia);
      end;
    except

    end;
  finally
    Serial.Free;
  end;
  //Json := Json.Replace('\r\n', '\n', [rfReplaceAll]);

  Body := TStringStream.Create(Json, TEncoding.UTF8);
  Stream := TMemoryStream.Create;
  try
    try
      Body.Seek(0, TSeekOrigin.soBeginning);
      IdHTTP.Post(PostTweetUrl_v2, Body, Stream);

      result := MemoryStreamToString(Stream);
      TFile.AppendAllText('Twitter.log', Text + ' - ' + result + SLineBreak, TEncoding.UTF8);

    except
      on e: exception do begin
        TFile.AppendAllText('HttpError.log',
          FormatDateTime('yyyy-mm-dd hh:nn:ss - ', Now) +
          'TTwitterBot2.DoPostTweet_v2: ' + e.Message + sLineBreak,
          TEncoding.UTF8);
        FHttpError := true;
      end;
    end;
  finally
    Body.Free;
    Stream.Free;
    if IdHTTP.Connected then
      IdHTTP.Disconnect;
  end;

  if FHttpError then
    ResetHTTP;

end;

end.
